import furl
from storages.backends import s3boto3


def boolean_str(s):
    return str(s).lower() in ["1", "true", "yes", "on"]


class S3Storage(s3boto3.S3Boto3Storage):
    def __init__(self, dsn):
        for needle in [
            ".s3-accesspoint-fips.dualstack.",
            ".s3-fips.dualstack.",
            ".s3-accesspoint.dualstack.",
            ".s3.dualstack.",
            ".s3-accesspoint.",
            ".s3-fips.",
            ".s3-",
            ".s3.",
        ]:
            bucket_name, marker, storage_host = dsn.host.partition(needle)
            if not marker:
                continue
            region_name = storage_host.split(".", 1)[0]
            if region_name == "amazonaws":
                region_name = None
            elif region_name in ["control", "control-fips"]:
                # Variants with account_id in the endpoint are unsupported
                continue
            storage_host = marker.lstrip(".") + storage_host
            break
        else:
            # Fall back to old static behavior
            endpoint = dsn.host.rsplit(".", 3)
            bucket_name = endpoint[0]
            storage_host = ".".join(endpoint[1:])
            region_name = (
                dsn.args.get("region_name", endpoint[1].partition("-")[2]) or None
            )

        location = str(dsn.path).lstrip("/")
        addressing_style = dsn.args.get("addressing_style")
        if not addressing_style:
            if "." in bucket_name:
                addressing_style = "path"
            else:
                addressing_style = "auto"

        if "url" in dsn.args:
            base_url = furl.furl(dsn.args.get("url"))
            url_protocol = "{}:".format(base_url.scheme)
            custom_domain = base_url.netloc
        else:
            url_protocol = "https:"
            custom_domain = dsn.args.get("domain")
            base_url = furl.furl()
            base_url.scheme = "https"
            base_url.host = custom_domain or dsn.host
            base_url.path = location.rstrip("/") + "/"

        super().__init__(
            access_key=dsn.username or "",
            secret_key=dsn.password or "",
            bucket_name=bucket_name,
            use_ssl=True,
            endpoint_url="https://{}/".format(storage_host),
            addressing_style=addressing_style,
            signature_version=dsn.args.get("auth", None),
            location=location,
            region_name=region_name,
            custom_domain=custom_domain,
            # TODO: Make the default `private` and explicitly set the ACL to
            #       `public-read` during provisioning
            object_parameters={"ACL": dsn.args.get("acl", "public-read")},
            # TODO: Support querystring_auth=True + custom_domain
            querystring_auth=boolean_str(dsn.args.get("qs_auth", "false")),
            url_protocol=url_protocol,
            # TODO: Enforce encryption everywhere. Check status on non-AWS
            #       providers.
            # encryption=True,
            max_memory_size=10 * 1024**2,  # Roll over to disk after 10 MB
        )

        self.base_url = str(base_url)



import base64
from urllib.parse import parse_qs

import furl
from storages.backends import azure_storage


class AzureStorageFile(azure_storage.AzureStorageFile):
    def __init__(self, name, mode, storage):
        super().__init__(name, mode, storage)
        if "w" not in mode:
            # Force early RAII-style exception if object does not exist
            if not storage.exists(name):
                raise IOError("File does not exist: %s" % name)


class AzureStorage(azure_storage.AzureStorage):
    mimetype_overrides = {
        ("application/x-tar", "gzip"): ("application/x-gtar", None),
    }

    def __init__(self, dsn):
        account_name = dsn.username
        credential = dsn.password
        credential += "=" * (-len(credential) % 4)

        if "url" in dsn.args:
            base_url = furl.furl(dsn.args.get("url"))
            secure_urls = base_url.scheme == "https"
            custom_domain = base_url.netloc
        else:
            secure_urls = True
            custom_domain = dsn.args.get("domain")
            base_url = furl.furl()
            base_url.scheme = "https"
            base_url.host = custom_domain or "{}.{}".format(account_name, dsn.host)

        container_name = str(dsn.path).strip("/")
        base_url.path = container_name + "/"

        super().__init__()
        self.account_name = account_name
        try:
            # SAS token
            sas_token = base64.b64decode(credential).decode("ascii")
        except UnicodeDecodeError:
            # Account key (binary data)
            self.account_key = credential
        else:
            if "sig" in parse_qs(sas_token):
                self.sas_token = sas_token
            else:
                # Account key (ascii)
                self.account_key = credential

        self.azure_container = container_name
        self.azure_ssl = secure_urls
        self.max_memory_size = 10 * 1024**2
        self.overwrite_files = True
        self.location = ""
        self.base_url = str(base_url)

    def url(self, name, expire=None, reset_query=True):
        url = super().url(name, expire)
        url = furl.furl(url)
        # TODO: How does this work when using a custom url
        #       with an additional domain?
        url.netloc = furl.furl(self.base_url).netloc

        if reset_query:
            # By default, reset the query parameters to remove SAS tokens from the URL.
            url.query = ""
        return str(url)

    def _open(self, name, mode="rb"):
        return AzureStorageFile(name, mode, self)

    def _get_content_settings_parameters(self, name, content=None):
        # Azure forwards the Content-Encoding header set during upload to
        # download request responses (causing transparent decompression by most
        # clients when it is set to ``gzip``).
        # The superclass uses the return value of ``mimetypes.guess_type``,
        # which is explicitly documented as not suitable for
        # content *transfer* encoding.
        #
        # NOTE: The content-encoding header should probably *never* be set by
        #       this storage backend.
        #
        # https://docs.python.org/3/library/mimetypes.html#mimetypes.guess_type
        #
        params = super()._get_content_settings_parameters(name, content=content)
        type_enc = (params["content_type"], params["content_encoding"])
        (
            params["content_type"],
            params["content_encoding"],
        ) = self.mimetype_overrides.get(type_enc, type_enc)
        return params



import pickle

from django.test import TestCase, override_settings
from django.db.migrations.serializer import serializer_factory

from django_storage_url import dsn_configured_storage


class DeconstructTestCase(TestCase):
    @override_settings(STORAGE_DSN="file:///test/media")
    def test_generated_class_module(self):
        storage = dsn_configured_storage("STORAGE_DSN")
        storage._setup()  # Storage is a LazyObject, make sure it is set up
        self.assertEqual(
            storage._wrapped.__module__, "django_storage_url.backends.file"
        )
        self.assertEqual(
            storage._wrapped.__class__.__module__,
            "django_storage_url.backends.file",
        )

    def test_not_implemented_deconstructible(self):
        storage = dsn_configured_storage("UNDEFINED_DSN")

        arg_string, arg_imports = serializer_factory(storage).serialize()
        self.assertEqual(
            arg_string,
            "django_storage_url.base.dsn_configured_storage('UNDEFINED_DSN')",
        )
        self.assertEqual(arg_imports, {"import django_storage_url.base"})

    @override_settings(STORAGE_DSN="file:///test/media")
    def test_defined_deconstructible(self):
        storage = dsn_configured_storage("STORAGE_DSN")

        arg_string, arg_imports = serializer_factory(storage).serialize()
        self.assertEqual(
            arg_string,
            "django_storage_url.base.dsn_configured_storage('STORAGE_DSN')",
        )
        self.assertEqual(arg_imports, {"import django_storage_url.base"})


class PickleTestCase(TestCase):
    @override_settings(STORAGE_DSN="file:///test/media")
    def test_pickle_lazy(self):
        storage = dsn_configured_storage("STORAGE_DSN")

        dumped = pickle.dumps(storage)
        loaded = pickle.loads(dumped)

        # Make sure that once unpickled it's still referring to the same object
        storage._setup()
        loaded._setup()

        assert loaded._wrapped.__dict__ == storage._wrapped.__dict__
        assert loaded._wrapped.__module__ == storage._wrapped.__module__
        assert loaded._wrapped.__class__.__name__ == storage._wrapped.__class__.__name__
        assert (
            loaded._wrapped.__class__.__bases__ == storage._wrapped.__class__.__bases__
        )

    @override_settings(STORAGE_DSN="file:///test/media")
    def test_pickle_wrapped(self):
        storage = dsn_configured_storage("STORAGE_DSN")
        storage._setup()

        dumped = pickle.dumps(storage)
        loaded = pickle.loads(dumped)

        # Make sure that once unpickled it's still referring to the same object
        loaded._setup()

        assert loaded._wrapped.__dict__ == storage._wrapped.__dict__
        assert loaded._wrapped.__module__ == storage._wrapped.__module__
        assert loaded._wrapped.__class__.__name__ == storage._wrapped.__class__.__name__
        assert (
            loaded._wrapped.__class__.__bases__ == storage._wrapped.__class__.__bases__
        )



from .backends import get_storage_class, register_storage_class
from .base import (
    dsn_configured_storage,
    dsn_configured_storage_class,
    get_storage,
)


# for __version__ see setup.py

__all__ = [
    "dsn_configured_storage",
    "dsn_configured_storage_class",
    "register_storage_class",
    "get_storage_class",
    "get_storage",
]
